/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.plugins.pmd.cpd;

import ch.hortis.sonar.service.MeasureKey;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.maven.plugin.logging.Log;
import org.sonar.plugins.api.maven.*;
import org.sonar.plugins.api.maven.model.MavenPom;
import org.sonar.plugins.api.maven.xml.*;
import org.sonar.plugins.api.metrics.CoreMetrics;
import org.w3c.dom.*;

import java.io.*;
import java.text.ParseException;
import java.util.*;

public class CpdMavenCollector extends AbstractJavaMavenCollector {

  public CpdMavenCollector(JavaMeasuresRecorder recorder, Log log) {
    super(recorder, log);
  }

  protected boolean shouldCollectIfNoSources() {
    return false;
  }

  public Class<? extends MavenPluginHandler> dependsOnMavenPlugin(MavenPom pom) {
    return CpdMavenPluginHandler.class;
  }

  public boolean shouldStopOnFailure() {
    return true;
  }

  public void collect(MavenPom pom) {
    File xmlFile = MavenCollectorUtils.findFileFromBuildDirectory(pom, "cpd.xml");
    collect(xmlFile, pom.getCompileSourceRoots());
  }


  protected void collect(File reportXmlFile, List<String> sourceDirs) {
    try {
      if (reportXmlFile != null && reportXmlFile.exists()) {
        XmlReportParser parser = new XmlReportParser();
        parser.parse(readXmlWithoutEncodingErrors(reportXmlFile));

        NodeList duplications = parser.executeXPathNodeList("/pmd-cpd/duplication");
        Map<String, ClassTempData> classData = new HashMap<String, ClassTempData>();
        for (int i = 0; i < duplications.getLength(); i++) {
          Element duplication = (Element) duplications.item(i);

          NodeList files = parser.executeXPathNodeList(duplication, "file");
          processFileMeasure(classData, (Element) files.item(0), duplication, sourceDirs);
          processFileMeasure(classData, (Element) files.item(1), duplication, sourceDirs);
        }

        for (ClassTempData data : classData.values()) {
          data.saveUsing(getMeasuresRecorder());
        }
      }
    } catch (ParseException e) {
      throw new XmlParserException(e);
    }
  }

  private String readXmlWithoutEncodingErrors(File file) {
    try {
      // First step : the file is read with system charset encoding. It should resolve the problem in most cases
      String xml = FileUtils.readFileToString(file);

      // second step : remove CDATA nodes that contain wrong characters. Those nodes are not needed by the collector.
      return removeCDataNodes(xml);

    } catch (IOException e) {
      throw new XmlParserException("can not read the file " + file.getAbsolutePath(), e);
    }
  }

  String removeCDataNodes(String xml) {
    String result = xml;
    String startNode = "<codefragment>";
    String endNode = "</codefragment>";
    String[] subs = StringUtils.substringsBetween(xml, startNode, endNode);
    if (subs != null) {
      for (String sub : subs) {
        result = StringUtils.remove(result, startNode + sub + endNode);
      }
    }
    return result;
  }

  private void processFileMeasure(Map<String, ClassTempData> fileContainer, Element fileEl, Element duplication, List<String> sourceDirs) throws ParseException {
    String filePath = fileEl.getAttribute("path").replace('\\', '/');
    String fileName = StringUtils.substringAfterLast(filePath, "/");
    fileName = StringUtils.substringBeforeLast(fileName, ".java");
    String packageName = StringUtils.substringBeforeLast(filePath, "/");
    if (sourceDirs != null) {
      for (String sourceDir : sourceDirs) {
        String source = sourceDir.replace('\\', '/');
        if (packageName.startsWith(source) || packageName.contains(source)) {
          if (!source.endsWith("/")) {
            source += "/";
          }
          packageName = StringUtils.substringAfter(packageName, source).replace('/', '.');
          packageName = packageName.replace('/', '.');

          cumulateClassMeasures(packageName, fileName, duplication, fileContainer);
        }
      }
    }
  }

  private void cumulateClassMeasures(String packageName, String fileName, Element duplication, Map<String, ClassTempData> fileContainer) throws ParseException {
    ClassTempData data = fileContainer.get(packageName + fileName);
    if (data == null) {
      data = new ClassTempData(packageName, fileName);
      fileContainer.put(packageName + fileName, data);
    }

    data.cumulate(1, MavenCollectorUtils.parseNumber(duplication.getAttribute("lines")), MavenCollectorUtils.parseNumber(duplication.getAttribute("tokens")));
  }

  private static class ClassTempData {
    protected double duplication;
    protected double duplicatedLines;
    protected double duplicatedTokens;
    protected String namespace;
    protected String filename;

    private ClassTempData(String namespace, String filename) {
      this.namespace = namespace;
      this.filename = filename;
    }

    protected void cumulate(double duplication, double duplicatedLines, double duplicatedTokens) {
      this.duplication += duplication;
      this.duplicatedLines += duplicatedLines;
      this.duplicatedTokens += duplicatedTokens;
    }

    protected void saveUsing(JavaMeasuresRecorder recorder) {
      //recorder.createClassMeasure(namespace, filename, false, new MeasureKey(MetricsRepository.DUPLICATIONS), duplication);
      recorder.createClassMeasure(namespace, filename, false, new MeasureKey(CoreMetrics.DUPLICATED_LINES), duplicatedLines);
      //recorder.createClassMeasure(namespace, filename, false, new MeasureKey(MetricsRepository.DUPLICATED_TOKENS), duplicatedTokens);
    }
  }
}
